@@ -129,12 +129,6 @@ static const int signals[] = {
 #ifdef SIGQUIT
 	SIGQUIT,
 #endif
-#ifdef SIGILL
-	SIGILL,
-#endif
-#ifdef SIGABRT
-	SIGABRT,
-#endif
 #ifdef SIGFPE
 	SIGFPE,
 #endif
@@ -351,33 +345,45 @@ static void pr_nomem(const char *msg)
  *  get_pid_cmdline
  * 	get process's /proc/pid/cmdline
  */
-static char *get_pid_cmdline(const pid_t id)
+static char *get_pid_cmdline(const pid_t pid)
 {
 	char buffer[BUFFER_SIZE];
 	char *ptr;
 	int fd;
-	ssize_t ret;
+	ssize_t ret = 0;
 
-	snprintf(buffer, sizeof(buffer), "/proc/%d/cmdline", id);
-
-	if ((fd = open(buffer, O_RDONLY)) < 0)
-		return strdup(CMD_UNKNOWN);
-
-	memset(buffer, 0, sizeof(buffer));
-	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0) {
+	*buffer = '\0';
+	(void)snprintf(buffer, sizeof(buffer), "/proc/%d/cmdline", pid);
+	if ((fd = open(buffer, O_RDONLY)) > -1) {
+		ret = read(fd, buffer, sizeof(buffer) - 1);
 		(void)close(fd);
-		return strdup(CMD_UNKNOWN);
+		if (ret > -1)
+			buffer[ret] = '\0';
 	}
-	(void)close(fd);
-
-	buffer[sizeof(buffer)-1] = '\0';
-	for (ptr = buffer; *ptr && (ptr < buffer + ret); ptr++) {
-		if (*ptr == ' ') {
-			*ptr = '\0';
-			break;
+	/*
+	 * No cmdline, could be a kernel thread, so get the comm
+	 * field instead
+	 */
+	if (!*buffer) {
+		(void)snprintf(buffer, sizeof(buffer), "/proc/%d/comm", pid);
+		if ((fd = open(buffer, O_RDONLY)) > -1) {
+			ret = read(fd, buffer, sizeof(buffer) - 1);
+			(void)close(fd);
+			if (ret > 0)
+				buffer[ret - 1] = '\0';  /* remove trailing \n */
 		}
 	}
 
+	if (ret < 1) {
+		strncpy(buffer, "<unknown>", sizeof(buffer));
+	} else {
+		for (ptr = buffer; *ptr && (ptr < buffer + ret); ptr++) {
+			if (*ptr == ' ') {
+				*ptr = '\0';
+				break;
+			}
+		}
+	}
 	return strdup(basename(buffer));
 }
 
@@ -402,25 +408,19 @@ static inline double timeval_double(const struct timeval *tv)
 }
 
 /*
- *  hash_pjw()
- *	Hash a string, from Aho, Sethi, Ullman, Compiling Techniques.
+ *  hash_djb2a()
+ *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
  */
-static unsigned long hash_pjw(const char *str, const pid_t pid)
+static uint32_t hash_djb2a(const char *str, const pid_t pid)
 {
-  	unsigned long h = pid;
+	register uint32_t hash = 5381 + pid;
+	register int c;
 
-	while (*str) {
-		unsigned long g;
-
-		h = (h << 4) + (*str);
-		if (0 != (g = h & 0xf0000000)) {
-			h = h ^ (g >> 24);
-			h = h ^ g;
-		}
-		str++;
+	while ((c = *str++)) {
+		/* (hash * 33) ^ c */
+		hash = ((hash << 5) + hash) ^ c;
 	}
-
-  	return h % TABLE_SIZE;
+	return hash % TABLE_SIZE;
 }
 
 /*
@@ -452,7 +452,7 @@ static proc_info_t *proc_info_get(const pid_t pid)
 				/* It's probably died, so no new process name */
 				free(cmdline);
 				return pi;
-			} 
+			}
 			/* Deemed "stale", so refresh stats */
 			free(pi->cmdline);
 			pi->cmdline = cmdline;
@@ -558,7 +558,7 @@ static void file_stat_free(void)
  */
 static file_stat_t *file_stat_get(const char *str, const pid_t pid)
 {
-	const unsigned long h = hash_pjw(str, pid);
+	const unsigned long h = hash_djb2a(str, pid);
 	file_stat_t *fs = file_stats[h];
 
 	while (fs) {
@@ -582,21 +582,41 @@ static file_stat_t *file_stat_get(const char *str, const pid_t pid)
 	return fs;
 }
 
+
+/*
+ *  filter_out()
+ *	return true if pathname is excluded and not included
+ *	- excludes takes higher priority over includes
+ */
+static bool filter_out(const char *pathname)
+{
+	pathname_t *pe;
+
+	/* Check if pathname is on exclude list */
+	for (pe = paths_exclude; pe; pe = pe->next) {
+		if (!strncmp(pe->pathname, pathname, pe->pathlen))
+			return true;
+	}
+	/* No include list, assume include all */
+	if (!paths_include)
+		return false;
+
+	/* Check if pathname is on the include list */
+	for (pe = paths_include; pe; pe = pe->next) {
+		if (!strncmp(pe->pathname, pathname, pe->pathlen))
+			return false;
+	}
+	return true;
+}
+
 /*
  *  mark()
- *	add a new fanotify mask if filename not in
- *	exclude list and increment the counter
+ *	add a new fanotify mask
  */
 static int mark(int fan_fd, const char *pathname, int *count)
 {
-	pathname_t *pe;
 	int ret;
 
-	/* Check if pathname is not on exclude list */
-	for (pe = paths_exclude; pe; pe = pe->next) {
-		if (!strncmp(pe->pathname, pathname, pe->pathlen))
-			return 0;
-	}
 	ret = fanotify_mark(fan_fd, FAN_MARK_ADD | FAN_MARK_MOUNT,
 		FAN_ACCESS| FAN_MODIFY | FAN_OPEN | FAN_CLOSE |
 		FAN_ONDIR | FAN_EVENT_ON_CHILD, AT_FDCWD, pathname);
@@ -614,34 +634,27 @@ static int mark(int fan_fd, const char *pathname, int *count)
 static int fnotify_event_init(void)
 {
 	int fan_fd, count = 0;
-	FILE* mounts;
-	struct mntent* mount;
 
 	if ((fan_fd = fanotify_init(0, 0)) < 0)
 		pr_error("cannot initialize fanotify");
 
-	if (paths_include) {
-		pathname_t *pi;
+	FILE* mounts;
+	struct mntent* mount;
 
-		for (pi = paths_include; pi; pi = pi->next) {
-			(void)mark(fan_fd, pi->pathname, &count);
-		}
-	} else {
-		/* No paths given, do all mount points */
-		if ((mounts = setmntent("/proc/self/mounts", "r")) == NULL) {
-			(void)close(fan_fd);
-			pr_error("setmntent cannot get mount points from /proc/self/mounts");
-		}
-		/*
-		 *  Gather all mounted file systems and monitor them
-		 */
-		while ((mount = getmntent(mounts)) != NULL)
-			(void)mark(fan_fd, mount->mnt_dir, &count);
+	/* No paths given, do all mount points */
+	if ((mounts = setmntent("/proc/self/mounts", "r")) == NULL) {
+		(void)close(fan_fd);
+		pr_error("setmntent cannot get mount points from /proc/self/mounts");
+	}
+	/*
+	 *  Gather all mounted file systems and monitor them
+	 */
+	while ((mount = getmntent(mounts)) != NULL)
+		(void)mark(fan_fd, mount->mnt_dir, &count);
 
-		if (endmntent(mounts) < 0) {
-			(void)close(fan_fd);
-			pr_error("endmntent failed");
-		}
+	if (endmntent(mounts) < 0) {
+		(void)close(fan_fd);
+		pr_error("endmntent failed");
 	}
 
 	/* This really should not happen, / is always mounted */
@@ -650,7 +663,6 @@ static int fnotify_event_init(void)
 		(void)close(fan_fd);
 		exit(EXIT_FAILURE);
 	}
-
 	return fan_fd;
 }
 
@@ -857,9 +869,11 @@ static int fnotify_event_add(const struct fanotify_event_metadata *metadata)
 
  	filename = fnotify_get_filename(-1, metadata->fd);
 	if (filename == NULL) {
-		(void)close(metadata->fd);
 		pr_error("allocating fnotify filename");
+		goto tidy;
 	}
+	if (filter_out(filename))
+		goto tidy;
 
 	fs = file_stat_get(filename, metadata->pid);
 	if (metadata->mask & FAN_OPEN) {
@@ -960,8 +974,8 @@ static void file_stat_dump(const double duration, const unsigned long top)
 		*ts = '\0';
 	}
 	printf("Total   Open  Close   Read  Write   PID  Process         %s%s\n",
-		opt_flags & OPT_INODE ? "Dev (Maj:Min) Inode" :
-		opt_flags & OPT_DEVICE ? "Dev (Maj:Min)" : "Pathname", ts);
+		(opt_flags & OPT_INODE) ? "Dev (Maj:Min) Inode" :
+		(opt_flags & OPT_DEVICE) ? "Dev (Maj:Min)" : "Pathname", ts);
 	for (j = 0; j < file_stats_size; j++) {
 		if (top && j <= top) {
 			char buf[5][32];
@@ -995,9 +1009,9 @@ static void file_stat_dump(const double duration, const unsigned long top)
  */
 static int parse_pid_list(char *arg)
 {
-	char *str, *token, *saveptr = NULL;
+	char *str, *token;
 
-	for (str = arg; (token = strtok_r(str, ",", &saveptr)) != NULL; str = NULL) {
+	for (str = arg; (token = strtok(str, ",")) != NULL; str = NULL) {
 		pid_t pid = 0;
 		char *name = NULL;
 		proc_info_t *p;
@@ -1046,10 +1060,10 @@ static int parse_pid_list(char *arg)
  */
 static int parse_pathname_list(char *arg, pathname_t **list)
 {
-	char *str, *token, *saveptr = NULL;
+	char *str, *token;
 	bool added = false;
 
-	for (str = arg; (token = strtok_r(str, ",", &saveptr)) != NULL; str = NULL) {
+	for (str = arg; (token = strtok(str, ",")) != NULL; str = NULL) {
 		pathname_t *p;
 
 		p = calloc(1, sizeof(pathname_t));
